<?php

namespace Drupal\knowledge\EventSubscriber;

use Drupal\Core\Config\ConfigCrudEvent;
use Drupal\Core\Config\ConfigEvents;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\Utility\Error;
use Psr\Log\LoggerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Creates or removes actions to match available waves.
 *
 * @package Drupal\knowledge\EventSubscriber
 */
class WaveEventSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * Drupal\Core\Entity\EntityStorageInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $storage;

  /**
   * The string translation service.
   *
   * @var \Drupal\Core\StringTranslation\TranslationInterface
   */
  protected $stringTranslation;

  /**
   * The logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Creates the event subscriber.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The action storage.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger service.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    TranslationInterface $string_translation,
    LoggerInterface $logger
    ) {
    $this->storage = $entity_type_manager->getStorage('action');
    $this->stringTranslation = $string_translation;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   *
   * @return array
   *   The event names to listen for, and the methods that should be executed.
   */
  public static function getSubscribedEvents() {
    return [
      ConfigEvents::SAVE => 'configSave',
    ];
  }

  /**
   * React to a config object being saved.
   *
   * @param \Drupal\Core\Config\ConfigCrudEvent $event
   *   Config crud event.
   */
  public function configSave(ConfigCrudEvent $event) {
    $config = $event->getConfig();
    $name = $config->getName();
    if ($name != 'field.storage.user.knowledge_wave') {
      return NULL;
    }

    $allowed_values = $config->get('settings.allowed_values');
    $orginal_values = $config->getOriginal('settings.allowed_values');
    $av = [];
    foreach ($allowed_values as $key => $value) {
      $av[$key] = $value['value'];
    }
    $ov = [];
    foreach ($orginal_values as $key => $value) {
      $ov[$key] = $value['value'];
    }

    $remove = array_diff($ov, $av);
    $add = array_diff($av, $ov);

    foreach ($remove as $int) {
      $action_id = "knowledge_set_user_wave_action.$int";
      $action = $this->storage->load($action_id);
      if ($action) {
        try {
          $action->delete();
        }
        catch (\Exception $e) {
          Error::logException($this->logger, $e);
        }

      }
    }

    foreach ($allowed_values as $item) {
      if (in_array($item['value'], $add)) {
        $action = $this->storage->create([
          'id' => "knowledge_set_user_wave_action.{$item['value']}",
          'label' => $this->t('Set wave to @label for the selected user(s)', ['@label' => $item['label']]),
          'type' => 'user',
          'plugin' => 'knowledge_set_user_wave_action',
          'configuration' => [
            'wave' => $item['value'],
          ],
        ]);
        try {
          $action->save();
        }
        catch (\Exception $e) {
          Error::logException($this->logger, $e);
        }
      }
    }
  }

}
