<?php

namespace Drupal\knowledge\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\jsource\Service\JsourceCaseStatisticsInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines a confirmation form to confirm deletion of something by id.
 */
class NodeStatisticsForm extends FormBase {

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * The data warehouse API client.
   *
   * @var \Drupal\jsource\Service\WarehouseClientInterface
   */
  protected $client;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->database = $container->get('database');
    $instance->time = $container->get('datetime.time');
    $instance->client = $container->get('jsource.warehouse_client');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $start_date = $form_state->getValue('start_date') ?? '03/07/2022';
    $end_date = $form_state->getValue('end_date') ?? '03/14/2022';
    $form = [];
    $form['start_date'] = [
      '#type' => 'date',
      '#title' => $this->t('Start Date'),
      '#required' => TRUE,
      '#default_value' => $start_date,
    ];
    $form['end_date'] = [
      '#type' => 'date',
      '#title' => $this->t('End Date'),
      '#required' => TRUE,
      '#default_value' => $end_date,
    ];

    $form['import_data'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Import Cases'),
    ];
    $form['generate_stats'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Generate Statistics'),
    ];
    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('GET'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    $generate_stats = $form_state->getValue('generate_stats');
    $start_date = $form_state->getValue('start_date');
    $start_date = strtotime($start_date);
    $end_date = $form_state->getValue('end_date');
    $end_date = strtotime($end_date);

    $operations = [];
    while ($start_date <= $end_date) {

      $date = date('Y-m-d', $start_date);
      if ($generate_stats) {
        $operations[] = [[$this, 'statistics'], [$date]];
      }
      $start_date = strtotime('+1 day', $start_date);
    }

    $batch = [
      'title' => $this->t('Syncing'),
      'operations' => $operations,
      'finished' => [$this, 'statisticsFinished'],
    ];
    batch_set($batch);
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() : string {
    return "case_statistics_import";
  }

  /**
   * Generates cases statistics for all people on said date.
   */
  public function statistics($date, &$context) {
    if (!isset($context['sandbox']['progress'])) {
      $user_ids = $this->database->select('users', 'u')
        ->fields('u', ['uid'])
        ->condition('uid', 1, '>')
        ->execute();

      $ids = [];
      foreach ($user_ids as $id) {
        $ids[] = $id->uid;
      }
      $context['sandbox']['ids'] = $ids;
      $context['sandbox']['max'] = count($ids);
      $context['sandbox']['progress'] = 0;
    }

    $user_id = array_pop($context['sandbox']['ids']);
    $context['sandbox']['progress'] += 1;
    $time = strtotime($date);
    $closed_count = $this->statClosedOn($user_id, $date);
    $fcr_count = $this->statClosedFcrOn($user_id, $date);
    $knowledge_count = $this->statClosedWithKnowledge($user_id, $date);

    if ($closed_count > 0) {
      $this->database->merge('jsource_case_statistics')
        ->key(
          ['name', 'start', 'end', 'entity_type', 'entity_id', 'period'],
          ['closed', $time, $time, 'user', $user_id, 1])
        ->fields([
          'value' => $closed_count,
          'processed' => $this->time->getRequestTime(),
        ])
        ->expression('value', $closed_count)
        ->expression('processed', $this->time->getRequestTime())
        ->execute();
    }
    else {
      $this->database->delete('jsource_case_statistics')
        ->condition('name', 'closed')
        ->condition('start', $time)
        ->condition('entity_type', 'user')
        ->condition('entity_id', $user_id)
        ->condition('period', 1)
        ->execute();
    }

    if ($fcr_count > 0) {
      $this->database->merge('jsource_case_statistics')
        ->key(
          ['name', 'start', 'end', 'entity_type', 'entity_id', 'period'],
          ['first_contact_resolution', $time, $time, 'user', $user_id, 1])
        ->fields([
          'value' => $fcr_count,
          'processed' => $this->time->getRequestTime(),
        ])
        ->expression('value', $fcr_count)
        ->expression('processed', $this->time->getRequestTime())
        ->execute();
    }
    else {
      $this->database->delete('jsource_case_statistics')
        ->condition('name', 'first_contact_resolution')
        ->condition('start', $time)
        ->condition('entity_type', 'user')
        ->condition('entity_id', $user_id)
        ->condition('period', 1)
        ->execute();
    }
    if ($knowledge_count > 0) {
      $this->database->merge('jsource_case_statistics')
        ->key(
          ['name', 'start', 'end', 'entity_type', 'entity_id', 'period'],
          ['closed_with_knowledge', $time, $time, 'user', $user_id, 1])
        ->fields([
          'value' => $knowledge_count,
          'processed' => $this->time->getRequestTime(),
        ])
        ->expression('value', $knowledge_count)
        ->expression('processed', $this->time->getRequestTime())
        ->execute();
    }
    else {
      $this->database->delete('jsource_case_statistics')
        ->condition('name', 'closed_with_knowledge')
        ->condition('start', $time)
        ->condition('entity_type', 'user')
        ->condition('entity_id', $user_id)
        ->condition('period', 1)
        ->execute();
    }

    // Is last of week.
    if (date('w', $time) == 6) {
      $this->statWeekly($user_id, $time);
    }
    // Is last of month.
    $last = date("Y-m-t", $time);
    $today = date("Y-m-d", $time);

    if ($today == $last) {
      $this->statMonthly($user_id, $time);
    }
    if ($context['sandbox']['max']) {
      $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
    }
    if ($context['finished'] == 1) {
      // User 0 / all users.
      $this->statAllUser($date);
    }
  }

  /**
   * Finish method for batch.
   */
  public function statisticsFinished($success, $results, $operations, $elapsed) {

  }

  /**
   * Gets cases closed on $date assigned to $person_id.
   */
  protected function getClosedOn($person_id, $date) {
    $options = [
      'assignedTo' => $person_id,
      'closedDate' => $date,
    ];
    $cases = $this->client->getCases($options);

    return $cases;
  }

  /**
   * Generates weekly stats for a user.
   */
  protected function statWeekly($uid, $time) {
    $start_date = strtotime('sunday last week', $time);
    $end_date = $time;

    $query = $this->database->select('jsource_case_statistics', 'jcs')
      ->fields('jcs', ['name', 'entity_type', 'entity_id']);
    $query
      ->condition('entity_type', 'user')
      ->condition('entity_id', $uid)
      ->condition('start', $start_date, '>=')
      ->condition('end', $end_date, '<=')
      ->condition('period', JsourceCaseStatisticsInterface::DAY)
      ->addExpression('SUM(value)', 'value');
    $records = $query
      ->groupBy('name')
      ->groupBy('entity_type')
      ->groupBy('entity_id')
      ->execute()->fetchAll();
    foreach ($records as $record) {
      $this->database->merge('jsource_case_statistics')
        ->key(
          ['name', 'start', 'end', 'entity_type', 'entity_id', 'period'],
          [
            $record->name,
            $start_date,
            $end_date,
            $record->entity_type,
            $record->entity_id,
            JsourceCaseStatisticsInterface::WEEK,
          ])
        ->fields([
          'value' => $record->value,
          'processed' => $this->time->getRequestTime(),
        ])
        ->expression('value', $record->value)
        ->expression('processed', $this->time->getRequestTime())
        ->execute();
    }
  }

  /**
   * Generates monthly stats for a user.
   */
  protected function statMonthly($uid, $time) {

    $start_date = date("Y-m-01", $time);

    $start_date = strtotime($start_date);

    $end_date = $time;

    $query = $this->database->select('jsource_case_statistics', 'jcs')
      ->fields('jcs', ['name', 'entity_type', 'entity_id']);
    $query
      ->condition('entity_type', 'user')
      ->condition('entity_id', $uid)
      ->condition('start', $start_date, '>=')
      ->condition('end', $end_date, '<=')
      ->condition('period', JsourceCaseStatisticsInterface::WEEK)
      ->addExpression('SUM(value)', 'value');
    $query
      ->groupBy('name')
      ->groupBy('entity_type')
      ->groupBy('entity_id');
    $records = $query->execute()->fetchAll();

    foreach ($records as $record) {
      $this->database->merge('jsource_case_statistics')
        ->key(
          ['name', 'start', 'end', 'entity_type', 'entity_id', 'period'],
          [
            $record->name,
            $start_date,
            $end_date,
            $record->entity_type,
            $record->entity_id,
            JsourceCaseStatisticsInterface::MONTH,
          ])
        ->fields([
          'value' => $record->value,
          'processed' => $this->time->getRequestTime(),
        ])
        ->expression('value', $record->value)
        ->expression('processed', $this->time->getRequestTime())
        ->execute();
    }
  }

  /**
   * SUM all users stats.
   */
  protected function statAllUser($date) {
    $uid = 1;
    $end_date = strtotime($date);

    $query = $this->database->select('jsource_case_statistics', 'jcs')
      ->fields('jcs', ['name', 'start', 'end', 'period']);
    $query
      ->condition('entity_type', 'user')
      ->condition('entity_id', $uid, '<>')
      ->condition('end', $end_date)
      ->addExpression('SUM(value)', 'value');
    $records = $query
      ->groupBy('name')
      ->groupBy('start')
      ->groupBy('end')
      ->groupBy('period')
      ->execute()->fetchAll();

    foreach ($records as $record) {
      $this->database->merge('jsource_case_statistics')
        ->key(
          ['name', 'start', 'end', 'entity_type', 'entity_id', 'period'],
          [
            $record->name,
            $record->start,
            $record->end,
            'user',
            $uid,
            $record->period,
          ])
        ->fields([
          'value' => $record->value,
          'processed' => $this->time->getRequestTime(),
        ])
        ->expression('value', $record->value)
        ->expression('processed', $this->time->getRequestTime())
        ->execute();
    }

  }

  /**
   * Get the number of cases closed on $date, assigned to $user_id.
   */
  protected function statClosedOn($user_id, $date) {
    $start_date = strtotime($date);
    $end_date = strtotime('+1 day', $start_date);

    $query = $this->database->select('jsource_case', 'jc');
    $query->join('crm_core_user_sync_relation', 'ccusr', 'jc.assigned_to = ccusr.individual_id');

    $count = $query
      ->condition('jc.status', 'Closed')
      ->condition('jc.closed', $start_date, '>=')
      ->condition('jc.closed', $end_date, '<')
      ->condition('ccusr.user_id', $user_id)
      ->countQuery()
      ->execute()
      ->fetchField();

    return $count;
  }

  /**
   * Get the number of cases closed on $date, assigned to $user_id.
   */
  protected function statClosedFcrOn($user_id, $date) {
    $start_date = strtotime($date);
    $end_date = strtotime('+1 day', $start_date);

    $query = $this->database->select('jsource_case', 'jc');
    $query->join('crm_core_user_sync_relation', 'ccusr', 'jc.assigned_to = ccusr.individual_id');
    $query->join('jsource_case__resolved_by_first_contact', 'jcfrbfc', 'jc.id = jcfrbfc.entity_id');

    $query
      ->condition('jc.status', 'Closed')
      ->condition('jcfrbfc.resolved_by_first_contact_value', 1)
      ->condition('jc.closed', $start_date, '>=')
      ->condition('jc.closed', $end_date, '<')
      ->condition('ccusr.user_id', $user_id);

    $count = $query
      ->countQuery()
      ->execute()
      ->fetchField();

    return $count;
  }

  /**
   * Get the number of cases closed on $date, assigned to $user_id.
   */
  protected function statClosedWithKnowledge($user_id, $date) {
    $start_date = strtotime($date);
    $end_date = strtotime('+1 day', $start_date);

    $query = $this->database->select('jsource_case', 'jc');
    $query->fields('jc', ['id']);
    $query->join('crm_core_user_sync_relation', 'ccusr', 'jc.assigned_to = ccusr.individual_id');
    $query->join('knowledge', 'k', 'jc.id = k.incident_id AND ccusr.user_id = k.uid');
    $query
      ->condition('k.status', 1)
      ->condition('jc.status', 'Closed')
      ->condition('jc.closed', $start_date, '>=')
      ->condition('jc.closed', $end_date, '<')
      ->condition('ccusr.user_id', $user_id)
      ->distinct();

    $count = $query
      ->countQuery()
      ->execute()
      ->fetchField();

    return $count;
  }

}
